unit AviHandler;

interface
Uses windows, classes, MMSystem, AVIFile32, Dialogs, Graphics;
Type
  TFourCC = array [0..3] of Char;

  TAviHandler = class
   private
    FFile : PAVIFile;
    FStream : PAVISTREAM;
    FStreamInfo	: TAVIStreamInfo;
    FCodecName : TFourCC;

    FFrameCnt, FFramesPerSecond : Integer;
    FProcessing : Boolean;
    FFileName : String;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Close; virtual;

    property FrameCnt : Integer read FFrameCnt;
    property Processing : Boolean read FProcessing;
    property Filename : String read FFileName;
    property Codec : TFourCC read FCodecName write FCodecName;
    property FPS : Integer read FFramesPerSecond write FFramesPerSecond;
  end;


  TAviPlayer = class (TAviHandler)
   private
    FFrame : PGetFrame;
    FTmpBmp : TBitmap;
    FSize : TPoint;
  protected
    function CopyFrame(Src : PBitmapInfoHeader; Dst: TBitmap): Boolean;
  public
    destructor Destroy; override;

    function Open(const FileName: String): Boolean;
    function GetFrame(Dst : TBitmap; FrameIndex: Integer): Boolean;
    procedure Close; override;
    property FrameSize : TPoint read FSize;
  end;

  TAviRecorder = class (TAviHandler)
  private
    FCompStream : PAVISTREAM;
    FBitmapInfo	: TBitmapInfoHeader;
    FKeyFrameEvery : Integer;
  public
    function StartRecord(Const FileName : String; Src : TBitmap;
                         Quality : Integer = 10000; KeyFrameEvery : Integer = -1): Boolean;
    function AddFrame(Src: TBitmap): Boolean;
    procedure Close; override;
  end;

function GetCodecList(Src : TBitmap; List: TStrings): Integer;
function ShowCodecConfig(ParendWnd : THandle; Codec : TFourCC): Boolean;

implementation

Var AviInitDone : Boolean;

function AviFailed(Const Msg : String; Res : HResult): Boolean;
begin
Result := (Res <> AVIERR_OK);
If (Result) then ShowMessage(Msg);
//Sys_RaiseLastError;
end;


{ TAviHandler }

constructor TAviHandler.Create;
begin
If (not AVIInitDone) then begin
  AVIFileInit; AVIInitDone := True;
end;
end;

procedure TAviHandler.Close;
Var refCount : Integer;
begin
FCodecName := ''; FFrameCnt := 0; FFramesPerSecond := 0; FFileName := '';

If (not FProcessing) then Exit;

If (FStream <> nil) then begin
  AVIStreamEndStreaming(FStream);
  AVIStreamRelease(FStream);
  FStream := nil;
end;

repeat
  refCount := AviFileRelease(FFile);
until refCount <= 0;
FFile := nil;
FProcessing := False;
end;

destructor TAviHandler.Destroy;
begin
Close;
end;


{ TAviPlayer }


destructor TAviPlayer.Destroy;
begin
FTmpBmp.Free;
Close;
end;


procedure TAviPlayer.Close;
begin
Int64(FSize) := 0;

If (FProcessing) then begin
  If (FFrame <> nil) then begin
    AVIStreamGetFrameClose(FFrame);
    FFrame := nil;
  end;
  If (FStream <> nil) then AVIStreamEndStreaming(FStream);
end;

Inherited Close;
end;


function TAviPlayer.Open(Const FileName : String): Boolean;
begin
Result := False;
Close;

If AviFailed( 'Failed to open AVI video file',
              AVIFileOpen(FFile, PChar(FileName), OF_READ, nil) ) then Exit;

If AviFailed( 'Failed to get AVI video stream',
               AVIFileGetStream(FFile, FStream, streamtypeVIDEO, 0) ) then Exit;
// receiving 1-st video stream from file

FillChar(FStreamInfo, SizeOf(FStreamInfo), 0);
If AviFailed( 'Failed to get AVI stream info',
               AVIStreamInfo(FStream, FStreamInfo, SizeOf(FStreamInfo)) ) then Exit;

Integer(FCodecName) := FStreamInfo.fccHandler;
With FStreamInfo.rectFrame do begin
  FSize.x := Right - Left; FSize.y := Bottom - Top;
end;
FFrameCnt := FStreamInfo.dwLength - FStreamInfo.dwStart;
If FStreamInfo.dwScale <> 0 then
  FFramesPerSecond := Round(FStreamInfo.dwRate / FStreamInfo.dwScale)
else
  FFramesPerSecond := 30;

If AviFailed( 'Failed to begin AVI streaming',
  AVIStreamBeginStreaming(FStream, FStreamInfo.dwStart, FStreamInfo.dwLength, 1000)) then Exit;//FStreamInfo.dwRate);

FFrame := AVIStreamGetFrameOpen(FStream, nil);
If (FFrame = nil) then begin
  AviFailed( 'Failed to open AVI frames', -1); Exit;
end;

FProcessing := True; FFileName := FileName;
Result := True;
end;


procedure SetBmpSize(Width, Height, Bpp : Integer; Dst : TBitmap);
Const
  MaxPFormat = 4;
  PFormat : array[1..MaxPFormat] of TPixelFormat =
    (pf8bit, pf16bit, pf24bit, pf32bit);
Var pf : TPixelFormat;
begin
pf := PFormat[Bpp div 8];
If (Dst.PixelFormat <> pf) or (Dst.Width <> Width) or
   (Dst.Height <> Height) then
begin
  Dst.Width  := Width;
  Dst.Height := Height;
  Dst.PixelFormat := pf;
end;
end;


procedure Copy8to24(Src : PBitmapInfoHeader; Dst : TBitmap);
Var x,y,h,Cnt,Gap : Integer;
    SrcPal, pc : PRGBQuad;
    pb : PByte;
    dc : PRGBTriple;
begin
Gap := (Src.biSizeImage div Src.biHeight) - Src.biWidth;
pb := PByte(Src); Inc(pb, SizeOf(TBitmapInfoHeader));
SrcPal := PRGBQuad(pb);
If Src.biClrUsed <> 0 then Cnt := Src.biClrUsed
                      else Cnt := 1 shl Src.biBitCount;
Inc(pb, SizeOf(TRGBQuad) * Cnt);
h := Src.biHeight-1;
For y:=0 to h do begin
  dc := Dst.ScanLine[h-y];
  For x:=0 to Src.biWidth-1 do begin
    pc := SrcPal; Inc(pc, pb^);
    dc^ := PRGBTriple(pc)^;
    Inc(dc); Inc(pb);
  end;
  Inc(pb, Gap);
end;
end;

procedure Copy24to24(Src : PBitmapInfoHeader; Dst : TBitmap);
Var Bits : PByte;
begin
Bits := PByte(Src); Inc(Bits, Src.biSize);
Move(Bits^, Dst.ScanLine[Src.biHeight-1]^, Src.biSizeImage);
end;

procedure CopyByGDI(Src : PBitmapInfoHeader; Dst : TBitmap);
Var Bits : PByte;
begin
Bits := PByte(Src); Inc(Bits, Src.biSize);
SetDIBitsToDevice(Dst.Canvas.Handle, 0, 0, Src.biWidth, Src.biHeight,
                  0, 0, 0, Src.biHeight, Bits, PBitmapInfo(Src)^, DIB_RGB_COLORS);
end;


function TAviPlayer.CopyFrame(Src : PBitmapInfoHeader; Dst : TBitmap): Boolean;
begin
Result := (Src.biBitCount in [8, 16, 24, 32]);
If (not Result) then Exit;

SetBmpSize(Src.biWidth, Src.biHeight, 24, Dst);
Case Src.biBitCount of
  8 : begin
        If (FTmpBmp = nil) then FTmpBmp := TBitmap.Create;
        SetBmpSize(Src.biWidth, Src.biHeight, 8, FTmpBmp);
        Copy8to24(Src, Dst);
      end;
  24 : Copy24to24(Src, Dst);
  else CopyByGDI(Src, Dst);
end;
end;


function TAviPlayer.GetFrame(Dst : TBitmap; FrameIndex : Integer): Boolean;
Var FramePtr : PBitmapInfoHeader;
begin
Result := False;
If (not FProcessing) then Exit;

FramePtr := AVIStreamGetFrame(FFrame, FrameIndex + FStreamInfo.dwStart);

Result := (FramePtr <> nil) and ( CopyFrame(FramePtr, Dst) );
end;



{ TAviRecorder }

// only for non-palettized bitmaps
function GetBmpInfo(Src: TBitmap; Out Info : TBitmapInfoHeader): Boolean;
begin
FillChar(Info, SizeOf(Info), 0);
Info.biSize := SizeOf(TBitmapInfoHeader);
Result := (GetDIBits(Src.Canvas.Handle, Src.Handle, 0, 0, nil, PBitmapInfo(@Info)^, DIB_RGB_COLORS) <> 0);
end;


function TAviRecorder.StartRecord(const FileName: String; Src: TBitmap;
                                  Quality : Integer = 10000;
                                  KeyFrameEvery : Integer = -1): Boolean;
var
  AVIComp : TAVICompressOptions;
begin
Result := False;
DeleteFile(PChar(FileName));

If AviFailed( 'Failed to create AVI video work file',
              AVIFileOpen(FFile, PChar(FileName),
              OF_WRITE or OF_CREATE, nil) ) then Exit;

If (DWord(FCodecName) = 0) then FCodecName := 'MSVC';
If (FFramesPerSecond = 0) then FFramesPerSecond := 30;

FillChar(FStreamInfo, SizeOf(FStreamInfo), 0);
with FStreamInfo do begin
  fccType := streamtypeVIDEO;
  fccHandler := DWord(FCodecName);
  dwRate := 1000000;
  dwScale := dwRate div FFramesPerSecond; // dwRate / dwScale = FPS
  dwQuality := Quality;
  dwSuggestedBufferSize := 0;
  rectFrame.Right := Src.Width;
  rectFrame.Bottom := Src.Height;
end;

if AviFailed( 'Failed to create AVI video stream',
               AVIFileCreateStream(FFile, FStream, FStreamInfo) ) then Exit;

GetBmpInfo(Src, FBitmapInfo);

// default - key frame every 5 seconds
If (KeyFrameEvery = -1) then FKeyFrameEvery := FFramesPerSecond * 5
                        else FKeyFrameEvery := KeyFrameEvery;

FillChar(AVIComp, SizeOf(AVIComp), 0);
with AVIComp do begin
  fccType := streamtypeVIDEO;
  fccHandler := DWord(FCodecName);
  dwKeyFrameEvery := FKeyFrameEvery;//1000;//Round(1000 / FrameTime);
  dwQuality := Quality;
  dwFlags := AVICOMPRESSF_KEYFRAMES or AVICOMPRESSF_VALID;
  lpFormat := @FBitmapInfo;
  cbFormat := FBitmapInfo.biSize;
end;

if AviFailed('Failed to create compressed stream',
  AVIMakeCompressedStream(FCompStream, FStream, AviComp, PClsID(nil)^)) then Exit;

if AviFailed('Failed to set AVI stream format',
              AVIStreamSetFormat(FCompStream, 0, @FBitmapInfo, FBitmapInfo.biSize)) then Exit;

FFrameCnt := 0; FProcessing := True;
Result := True; FFileName := FileName;
end;

// all bitmaps must have the same format
function TAviRecorder.AddFrame(Src: TBitmap): Boolean;
Var Samples_Written, Bytes_Written, flag : Integer;
    Ptr : PByte;
//    bInfo : TBitmapInfoHeader;
begin
Result := False;
If (not FProcessing) then Exit;

If (FFrameCnt mod FKeyFrameEvery) = 0 then
  flag := AVIIF_KEYFRAME
else
  flag := 0;
 // seems that no matter what is here;
 // only AVIComp.dwKeyFrameEvery has meaning

//GetBmpInfo(Src, bInfo);

Ptr := Src.ScanLine[Src.Height-1];
if AviFailed('Failed to add frame to AVI',
             AVIStreamWrite(FCompStream, FFrameCnt, 1, Ptr,
                            FBitmapInfo.biSizeImage, flag,
                            @Samples_Written, @Bytes_Written)) then Exit;
Inc(FFrameCnt);
Result := True;
end;


procedure TAviRecorder.Close;
begin
If (FCompStream <> nil) then begin
  AVIStreamRelease(FCompStream);
  FCompStream := nil;
end;
Inherited Close;
end;




Const
  ICTYPE_VIDEO = $63646976; {vidc}
  ICMODE_COMPRESS = 1;
  ICMODE_QUERY = 4;

  ICM_USER = (DRV_USER + $0000);
  ICM_RESERVED_LOW = (DRV_USER + $1000);
  ICM_RESERVED_HIGH = (DRV_USER + $2000);
  ICM_RESERVED = ICM_RESERVED_LOW;
  ICM_COMPRESS_QUERY = (ICM_USER + 6); // query support for compress
  ICM_CONFIGURE = (ICM_RESERVED + 10); // show the configure dialog
  ICMF_CONFIGURE_QUERY = $00000001;

Type
  PICINFO = ^TICINFO;
  TICINFO = packed record
    dwSize: DWord; // sizeof(ICINFO)
    fccType: DWord; // compressor type     'vidc' 'audc'
    fccHandler: DWord; // compressor sub-type 'rle ' 'jpeg' 'pcm '
    dwFlags: DWord; // flags LOWORD is type specific
    dwVersion: DWord; // version of the driver
    dwVersionICM: DWord; // version of the ICM used
    // under Win32, the driver always returns UNICODE strings.
    szName: array[0..15] of WChar; // short name
    szDescription: array[0..127] of WChar; // DWORD name
    szDriver: array[0..127] of WChar; // driver that contains compressor
  end;


function ICInfo(fccType, fccHandler: DWord; lpicinfo: PICINFO): BOOL; stdcall; external 'MSVFW32.DLL';

function ICOpen(fccType, fccHandler: DWord; wMode: UINT): THandle; stdcall; external 'MSVFW32.DLL';

function ICSendMessage(hic: THandle; Msg: UINT; dw1, dw2: DWord): DWord; stdcall; external 'MSVFW32.DLL';

function ICCompressQuery(hic: THandle; lpbiInput, lpbiOutput: PBitmapInfoHeader): DWord;
begin
  Result := ICSendMessage(hic, ICM_COMPRESS_QUERY, DWord(lpbiInput), DWord(lpbiOutput));
end;

function ICGetInfo(hic: THandle; PICINFO: PICINFO; cb: DWord): DWord; stdcall; external 'MSVFW32.DLL';

function ICClose(hic: THandle): DWord; stdcall; external 'MSVFW32.DLL';


function GetCodecList(Src: TBitmap; List: TStrings): Integer;
var
  ii: TICINFO;
  i: DWord;
  ic: THandle;
  BitmapInfoHeader: TBitmapInfoHeader;
begin
List.Clear;
List.Add('DIB Uncompessed bitmap set');

GetBmpInfo(Src, BitmapInfoHeader);
FillChar(ii, SizeOf(ii), 0);
ii.dwSize := SizeOf(TICINFO);
i := 0;
While ICInfo(ICTYPE_VIDEO, i, @ii) do begin
  ic := ICOpen(ICTYPE_VIDEO, ii.fccHandler, ICMODE_QUERY);
  If ic <> 0 then begin
    If ICCompressQuery(ic, @BitmapInfoHeader, nil) = 0 then begin
      ICGetInfo(ic, @ii, SizeOf(ii));
      List.Add( TFourCC(ii.fccHandler) + ' ' + WideCharToString(ii.szDescription) );
    end;
    ICClose(ic);
  end;
  Inc(i);
end;
Result := List.Count;
end;

function ShowCodecConfig(ParendWnd : THandle; Codec : TFourCC): Boolean;
var
  ic: THandle;
begin
Result := False;
ic := ICOpen(ICTYPE_VIDEO, Integer(Codec), ICMODE_QUERY);
If ic <> 0 then begin
  Result := (ICSendMessage(ic, ICM_CONFIGURE, ParendWnd, 0) = 0);
  ICClose(ic);
end;
end;


initialization

finalization
If AVIInitDone then begin AviFileExit; AVIInitDone := False; end;

end.
